/*
 * Icon Task Manager
 *
 * Copyright 2011 Craig Drummond <craig.p.drummond@gmail.com>
 *
 * ----
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "dockitem.h"
#include "dockitemadaptor.h"
#include "dockmanager.h"
#include "abstracttaskitem.h"
#include "config.h"
#include <KDE/KDesktopFile>
#include <QtGui/QAction>
#include <QtCore/QFile>
#include <QtCore/QTimer>

static qulonglong itemCount=0;

DockItem::DockItem(const KUrl &desktopFile)
        : m_url(desktopFile)
        , m_timer(0)
        , m_progress(-1)
        , m_menuIdCount(0)
{
    new DockItemAdaptor(this);
    m_path=QLatin1String("/net/launchpad/DockManager/Item")+QString().setNum(itemCount++);
    QDBusConnection::sessionBus().registerObject(m_path, this);
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG << m_url;
#endif
}

DockItem::~DockItem()
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG << m_url;
#endif
    foreach(AbstractTaskItem *i, m_tasks) {
        i->setDockItem(0);
    }

    QDBusConnection::sessionBus().unregisterObject(m_path, QDBusConnection::UnregisterTree);
}

QString DockItem::DesktopFile() const
{
    return m_url.toLocalFile();
}

QString DockItem::Uri() const
{
    return m_url.url();
}

QString DockItem::name() const
{
    if(m_name.isEmpty() && m_url.isLocalFile() && KDesktopFile::isDesktopFile(m_url.toLocalFile())) {
        m_name = KDesktopFile(m_url.toLocalFile()).readName();
    }
    
    return m_name;
}

QList<QAction *> DockItem::menu() const
{
    return m_menu.values();
}

unsigned int DockItem::AddMenuItem(QMap<QString, QVariant> hints)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG << m_url << hints;
#endif
    if(calledFromDBus()) {
        DockManager::self()->itemService(this, message().service());
    }

    QString label,
            iconName,
            iconFile,
            container;

    /*if(hints.contains("uri")) {
    } else*/ {
        label=hints["label"].toString();
        iconName=hints["icon-name"].toString();
        iconFile=hints["icon-file"].toString();
        container=hints["container-title"].toString();
    }

    unsigned int id=m_menuIdCount++;
    QAction *action=iconFile.isEmpty() && iconName.isEmpty()
                    ? new QAction(label, this)
                    : new QAction(QIcon(iconFile.isEmpty() ? iconName : iconFile), label, this);
    connect(action, SIGNAL(triggered()), this, SLOT(menuActivated()));
    action->setData(id);
    action->setProperty("action-group-name", container);
    m_menu.insert(id, action);
    return id;
}

void DockItem::RemoveMenuItem(unsigned int id)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG << m_url  << id;
#endif
    if(calledFromDBus()) {
        DockManager::self()->itemService(this, message().service());
    }

    if(m_menu.contains(id)) {
        disconnect(m_menu[id], SIGNAL(triggered()), this, SLOT(menuActivated()));
        m_menu.remove(id);
    }
}

void DockItem::menuActivated()
{
    QObject *s=sender();
    if(s && qobject_cast<QAction *>(s)) {
        QAction *item=static_cast<QAction *>(s);
        emit MenuItemActivated(item->data().toUInt());
    }
}

void DockItem::UpdateDockItem(QMap<QString, QVariant> hints)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG << m_url << hints;
#endif
    if(calledFromDBus()) {
        DockManager::self()->itemService(this, message().service());
    }

    QMap<QString, QVariant>::ConstIterator it(hints.constBegin()),
                                           end(hints.constEnd());
    int                                    updated=0;
      
    for(; it!=end; ++it) {
        if(it.key()=="badge") {
            QString badge=it.value().toString();
            if(badge!=m_badge) {
                m_badge=badge;
                updated++;
            }
        } else if(it.key()=="progress") {
            int prog=it.value().toInt();
            if(prog!=m_progress) {
                m_progress=prog;
                updated++;
            }
        } else if(it.key()=="icon-file") {
            QString icon=it.value().toString();
            if(icon!=m_iconName) {
                m_iconName=icon;
                m_icon=m_iconName.isEmpty() || !QFile::exists(m_iconName) ? QIcon() : QIcon(m_iconName);
                updated++;
            }
        }
    }
    
    if(updated) {
        foreach(AbstractTaskItem *i, m_tasks) {
            i->dockItemUpdated();
        }
    }
}

void DockItem::registerTask(AbstractTaskItem *item)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG << m_url << (void *)item << item->text();
#endif
    m_tasks.insert(item);
    item->setDockItem(this);
    if(!m_iconName.isEmpty() || (m_progress>=0 && m_progress<=100)) {
        item->dockItemUpdated();
    }
    if(m_timer) {
        m_timer->stop();
    }
}

void DockItem::unregisterTask(AbstractTaskItem *item)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG << m_url << (void *)item << item->text();
#endif
    m_tasks.remove(item);
    
    if(0==m_tasks.count()) {
        // No current tasks, so set off timer. If nothing registers then we are no longer used...
        if(!m_timer) {
            m_timer=new QTimer(this);
            connect(m_timer, SIGNAL(timeout()), this, SLOT(check()));
        }
        m_timer->start(500);
    }
}

void DockItem::reset()
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_DM_DBUG;
#endif
    bool updated=!m_badge.isEmpty() || !m_iconName.isEmpty() || (m_progress>=0 && m_progress<=100);
        
    m_badge=QString();
    m_iconName=QString();
    m_icon=QIcon();
    m_progress=-1;

    foreach(QAction *mnu, m_menu.values()) {
        mnu->deleteLater();
    }
    
    m_menu.clear();

    if(updated) {
        foreach(AbstractTaskItem *i, m_tasks) {
            i->dockItemUpdated();
        }
    }
}

void DockItem::check()
{
    if(0==m_tasks.count()) {
        DockManager::self()->remove(this);
    }
}

#include "dockitem.moc"
