/*****************************************************************

Copyright 2008 Christian Mollekopf <chrigi_1@hotmail.com>
Copyright (C) 2011 Craig Drummond <craig.p.drummond@gmail.com> 

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#include "groupmanager.h"

#include <QBuffer>
#include <QList>
#include <QStack>
#include <QTimer>
#include <QUuid>
#include <QFile>

#include <KDebug>
#include <KSycoca>

#include "abstractsortingstrategy.h"
#include "startup.h"
#include "task.h"
#include "taskitem.h"
#include "taskgroup.h"
#include "taskmanager.h"
#include "strategies/alphasortingstrategy.h"
#include "strategies/desktopsortingstrategy.h"
#include "strategies/programgroupingstrategy.h"
#include "strategies/manualgroupingstrategy.h"
#include "strategies/kustodiangroupingstrategy.h"
#include "strategies/manualsortingstrategy.h"
#include "launcheritem.h"
#include "launcherconfig.h"
#ifdef ICON_ENABLE_DBUS_DBUG
#include <QTextDocument>
#endif

namespace TaskManager
{

class GroupManagerPrivate
{
public:
    GroupManagerPrivate(GroupManager *manager)
        : q(manager),
          sortingStrategy(GroupManager::NoSorting),
          groupingStrategy(GroupManager::NoGrouping),
          lastGroupingStrategy(GroupManager::NoGrouping),
          abstractGroupingStrategy(0),
          abstractSortingStrategy(0),
          currentScreen(-1),
          groupIsFullLimit(0),
          showOnlyCurrentDesktop(false),
          showOnlyCurrentActivity(false),
          showOnlyCurrentScreen(false),
          showOnlyMinimized(false),
          onlyGroupWhenFull(false),
          changingGroupingStrategy(false),
          readingLauncherConfig(false),
          separateLaunchers(true),
          groupAttention(false),
          launchersLocked(false)
    {
    }

    /** reload all tasks from TaskManager */
    void reloadTasks();
    void actuallyReloadTasks();

    /**
    * Keep track of changes in Taskmanager
    */
    void currentDesktopChanged(int);
    void currentActivityChanged(QString);
    void taskChanged(TaskPtr, ::TaskManager::TaskChanges);
    void checkScreenChange();
    void taskDestroyed(QObject *item);
    void startupItemDestroyed(AbstractGroupableItem *);
    void checkIfFull();
    void actuallyCheckIfFull();
    bool addTask(TaskPtr);
    void removeTask(TaskPtr);
    void addStartup(StartupPtr);
    void removeStartup(StartupPtr);
    void sycocaChanged(const QStringList &types);
    void launcherVisibilityChange();
    void checkLauncherVisibility(LauncherItem *launcher);
    void saveLauncherConfig();
    void saveLauncherConfig(KConfigGroup &cg);
    void readLauncherConfig();
    void readLauncherConfig(KConfigGroup &cg);
    KConfigGroup launcherConfig(const KConfigGroup &config = KConfigGroup());

    TaskGroup *currentRootGroup();

    GroupManager *q;
    QHash<StartupPtr, TaskItem*> startupList;
    GroupManager::TaskSortingStrategy sortingStrategy;
    GroupManager::TaskGroupingStrategy groupingStrategy;
    GroupManager::TaskGroupingStrategy lastGroupingStrategy;
    AbstractGroupingStrategy *abstractGroupingStrategy;
    AbstractSortingStrategy *abstractSortingStrategy;
    int currentScreen;
    QTimer screenTimer;
    QTimer reloadTimer;
    QTimer checkIfFullTimer;
    QSet<Task *> geometryTasks;
    int groupIsFullLimit;
    QUuid configToken;

    QHash<QString, QHash<int, TaskGroup*> > rootGroups; //container for groups
    QHash<KUrl, LauncherItem *> launchers;
    QList<Launcher> launcherSort;
    int currentDesktop;
    QString currentActivity;

    bool showOnlyCurrentDesktop : 1;
    bool showOnlyCurrentActivity : 1;
    bool showOnlyCurrentScreen : 1;
    bool showOnlyMinimized : 1;
    bool onlyGroupWhenFull : 1;
    bool changingGroupingStrategy : 1;
    bool readingLauncherConfig : 1;
    bool separateLaunchers : 1;
    bool groupAttention : 1;
    bool launchersLocked : 1;
};
    
GroupManager::GroupManager(QObject *parent)
    : QObject(parent),
      d(new GroupManagerPrivate(this))
{
    connect(TaskManager::self(), SIGNAL(taskAdded(TaskPtr)), this, SLOT(addTask(TaskPtr)));
    connect(TaskManager::self(), SIGNAL(taskRemoved(TaskPtr)), this, SLOT(removeTask(TaskPtr)));
    connect(TaskManager::self(), SIGNAL(startupAdded(StartupPtr)), this, SLOT(addStartup(StartupPtr)));
    connect(TaskManager::self(), SIGNAL(startupRemoved(StartupPtr)), this, SLOT(removeStartup(StartupPtr)));
    connect(KSycoca::self(), SIGNAL(databaseChanged(QStringList)), this, SLOT(sycocaChanged(const QStringList &)));

    d->currentDesktop = TaskManager::self()->currentDesktop();
    d->currentActivity = TaskManager::self()->currentActivity();

    d->rootGroups[d->currentActivity][d->currentDesktop] = new TaskGroup(this, "RootGroup", Qt::transparent);

    d->reloadTimer.setSingleShot(true);
    d->reloadTimer.setInterval(0);
    connect(&d->reloadTimer, SIGNAL(timeout()), this, SLOT(actuallyReloadTasks()));

    d->screenTimer.setSingleShot(true);
    d->screenTimer.setInterval(100);
    connect(&d->screenTimer, SIGNAL(timeout()), this, SLOT(checkScreenChange()));

    d->checkIfFullTimer.setSingleShot(true);
    d->checkIfFullTimer.setInterval(0);
    connect(&d->checkIfFullTimer, SIGNAL(timeout()), this, SLOT(actuallyCheckIfFull()));
}

GroupManager::~GroupManager()
{
    TaskManager::self()->setTrackGeometry(false, d->configToken);
    delete d->abstractSortingStrategy;
    delete d->abstractGroupingStrategy;
    delete d;
}

TaskGroup *GroupManagerPrivate::currentRootGroup()
{
    return rootGroups[currentActivity][currentDesktop];
}

void GroupManagerPrivate::reloadTasks()
{
    reloadTimer.start();
}

void GroupManagerPrivate::actuallyReloadTasks()
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG;
#endif
    //kDebug() << "number of tasks available " << TaskManager::self()->tasks().size();
    QHash<WId, TaskPtr> taskList = TaskManager::self()->tasks();
    QMutableHashIterator<WId, TaskPtr> it(taskList);

    while (it.hasNext()) {
        it.next();

        if (addTask(it.value())) {
            //kDebug() << "task added " << it.value()->visibleName();
            it.remove();
        }
    }

    // Remove what remains
    it.toFront();
    while (it.hasNext()) {
        it.next();
        removeTask(it.value());
    }

    emit q->reload();
}

void GroupManagerPrivate::addStartup(StartupPtr task)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << startupList.contains(task);
#endif
    //kDebug();
    if (!startupList.contains(task)) {
        TaskItem *item = new TaskItem(q, task);
        startupList.insert(task, item);
        currentRootGroup()->add(item);
        QObject::connect(item, SIGNAL(destroyed(AbstractGroupableItem*)),
                         q, SLOT(startupItemDestroyed(AbstractGroupableItem*)));
    }
}

void GroupManagerPrivate::removeStartup(StartupPtr task)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << startupList.contains(task);
#endif
    //kDebug();
    if (!startupList.contains(task)) {
        kDebug() << "invalid startup task";
        return;
    }

    TaskItem *item = startupList.take(task);
    if (item->parentGroup()) {
        item->parentGroup()->remove(item);
    }
    item->setTaskPointer(TaskPtr());
}

bool GroupManagerPrivate::addTask(TaskPtr task)
{
    #ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << task->visibleName() << task->className();
    #endif
    //kDebug();
    /* kDebug() << task->visibleName()
             << task->visibleNameWithState()
             << task->name()
             << task->className()
             << task->classClass(); */

    bool skip = false;
    if (!task->showInTaskbar()) {
        //kDebug() << "Do not show in taskbar";
        #ifdef ICON_ENABLE_DBUS_DBUG
        ICN_LIB_STD_DBUG << "Skipping !task->showInTaskbar()";
        #endif
        skip = true;
    }

    if (showOnlyCurrentScreen && !task->isOnScreen(currentScreen)) {
        //kDebug() << "Not on this screen and showOnlyCurrentScreen";
        #ifdef ICON_ENABLE_DBUS_DBUG
        ICN_LIB_STD_DBUG << "Skipping !task->showInTaskbar()";
        #endif
        skip = true;
    }

    // Should the Task be displayed ? We always display if attention is demaded
    if (!task->demandsAttention()) {
        // As the Task doesn't demand attention
        // go through all filters whether the task should be displayed or not
        if (showOnlyCurrentDesktop && !task->isOnCurrentDesktop()) {
            /* kDebug() << "Not on this desktop and showOnlyCurrentDesktop"
                     << KWindowSystem::currentDesktop() << task->desktop(); */
            #ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_STD_DBUG << "Skipping showOnlyCurrentDesktop && !task->isOnCurrentDesktop()";
            #endif
            skip = true;
        }

        if (showOnlyCurrentActivity && !task->isOnCurrentActivity()) {
            /* kDebug() << "Not on this desktop and showOnlyCurrentActivity"
                     << KWindowSystem::currentActivity() << task->desktop(); */
            #ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_STD_DBUG << "Skipping showOnlyCurrentActivity && !task->isOnCurrentActivity()";
            #endif
            skip = true;
        }

        if (showOnlyMinimized && !task->isMinimized()) {
            //kDebug() << "Not minimized and only showing minimized";
            #ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_STD_DBUG << "Skipping showOnlyMinimized && !task->isMinimized()";
            #endif
            skip = true;
        }

        NET::WindowType type = task->info().windowType(NET::NormalMask | NET::DialogMask |
                                                    NET::OverrideMask | NET::UtilityMask);
        if (type == NET::Utility) {
            //kDebug() << "skipping utility window" << task->name();
            #ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_STD_DBUG << "Skipping NET::Utility";
            #endif
            skip = true;
        }

            //TODO: should we check for transiency? if so the following code can detect it.
        /*
            QHash <TaskPtr, TaskItem*>::iterator it = d->itemList.begin();

            while (it != d->itemList.end()) {
                TaskItem *item = it.value();
                if (item->task()->hasTransient(task->window())) {
                    kDebug() << "TRANSIENT TRANSIENT TRANSIENT!";
                    return flase;
                }
                ++it;
            }
        */
    }

    //Ok the Task should be displayed
    TaskItem *item = qobject_cast<TaskItem*>(currentRootGroup()->getMemberByWId(task->window()));
    #ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << "Window:" << task->window() << "skip:" << skip << "task:" << (void *)item;
    #endif
    if (!item || skip) {
        TaskItem *startupItem = 0;
        QHash<StartupPtr, TaskItem *>::iterator it = startupList.begin();
        QHash<StartupPtr, TaskItem *>::iterator itEnd = startupList.end();
        while (it != itEnd) {
            if (it.key()->matchesWindow(task->window())) {
                //kDebug() << "startup task found";
                item = startupItem = it.value();
                #ifdef ICON_ENABLE_DBUS_DBUG
                ICN_LIB_STD_DBUG << "Window:" << task->window() << "startup task found:" << (void *)item;
                #endif
                startupList.erase(it);
                QObject::disconnect(item, 0, q, 0);
                if (!skip) {
                    item->setTaskPointer(task);
                }
                break;
            }
            ++it;
        }

        // if we are to skip because we don't display, we simply delete the startup related to it
        if (skip) {
            delete startupItem;
            #ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_STD_DBUG << "Skipping?";
            #endif
            return false;
        }

        if (!item) {
            item = new TaskItem(q, task);
            #ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_STD_DBUG << "Window:" << task->window() << "created new task:" << (void *)item;
            #endif
        }

        QObject::connect(task.data(), SIGNAL(destroyed(QObject*)),
                         q, SLOT(taskDestroyed(QObject*)));

        foreach (LauncherItem *launcher, launchers) {
            launcher->associateItemIfMatches(item);
        }
    }

    //Find a fitting group for the task with GroupingStrategies
    if (abstractGroupingStrategy && (groupAttention || !task->demandsAttention())) { //do not group attention tasks
        abstractGroupingStrategy->handleItem(item);
    } else {
        #ifdef ICON_ENABLE_DBUS_DBUG
        ICN_LIB_STD_DBUG << "strategey:" << (void *)abstractGroupingStrategy << (int)(abstractGroupingStrategy ? abstractGroupingStrategy->type() : -1) <<  "groupAttention:" << groupAttention << "att" << task->demandsAttention();
        #endif
        currentRootGroup()->add(item);
    }

    geometryTasks.insert(task.data());
    return true;
}


void GroupManagerPrivate::removeTask(TaskPtr task)
{
#ifdef ICON_ENABLE_DBUS_DBUG
     ICN_LIB_STD_DBUG << task->visibleName();
#endif
    //kDebug() << "remove: " << task->visibleName();
    geometryTasks.remove(task.data());

    AbstractGroupableItem *item = currentRootGroup()->getMemberByWId(task->window());
    if (!item) {
        // this can happen if the window hasn't been caught previously,
        // of it it is an ignored type such as a NET::Utility type window
        //kDebug() << "invalid item";
        return;
    }

    foreach (LauncherItem *launcher, launchers) {
        launcher->removeItemIfAssociated(item);
    }

    if (item->parentGroup()) {
        item->parentGroup()->remove(item);
    }

    //the item must exist as long as the TaskPtr does because of activate calls so don't delete the item here, it will delete itself.
}

void GroupManagerPrivate::taskDestroyed(QObject *item)
{
    Task *task = static_cast<Task*>(item);
    if (task) {
        geometryTasks.remove(task);
    }
}

void GroupManagerPrivate::startupItemDestroyed(AbstractGroupableItem *item)
{
    TaskItem *taskItem = static_cast<TaskItem*>(item);
    startupList.remove(startupList.key(taskItem));
    geometryTasks.remove(taskItem->task().data());
}

bool GroupManager::manualGroupingRequest(AbstractGroupableItem* item, TaskGroup* groupItem)
{
    if (d->abstractGroupingStrategy) {
        return d->abstractGroupingStrategy->manualGroupingRequest(item, groupItem);
    }
    return false;
}

bool GroupManager::manualGroupingRequest(ItemList items)
{
    if (d->abstractGroupingStrategy) {
            return d->abstractGroupingStrategy->manualGroupingRequest(items);
    }
    return false;
}

bool GroupManager::manualSortingRequest(AbstractGroupableItem* taskItem, int newIndex)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_DBUG << taskItem->name() << newIndex;
#endif
    if (d->abstractSortingStrategy && 
        (!launchersLocked() || separateLaunchers() || newIndex>=launcherCount())) {
        return d->abstractSortingStrategy->manualSortingRequest(taskItem, newIndex);
    }
    return false;
}

GroupPtr GroupManager::rootGroup() const
{
    return d->currentRootGroup();
}

void GroupManagerPrivate::currentActivityChanged(QString newActivity)
{
    if (!showOnlyCurrentActivity || currentActivity == newActivity) {
        return;
    }

    if (!rootGroups.contains(newActivity) || !rootGroups.value(newActivity).contains(currentDesktop)) {
        kDebug() << "created new desk group";
        rootGroups[newActivity][currentDesktop] = new TaskGroup(q, "RootGroup", Qt::transparent);
        if (abstractSortingStrategy) {
            abstractSortingStrategy->handleGroup(rootGroups[newActivity][currentDesktop]);
        }
    }

    if (onlyGroupWhenFull) {
        QObject::disconnect(currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
        QObject::disconnect(currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
    }

    currentActivity = newActivity;

    foreach (LauncherItem *item, launchers) {
        if (item->shouldShow(q)) {
            rootGroups[currentActivity][currentDesktop]->add(item);
        } else {
            rootGroups[currentActivity][currentDesktop]->remove(item);
        }
    }

    if (onlyGroupWhenFull) {
        QObject::connect(currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
        QObject::connect(currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
    }
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << "Call reload";
#endif
    actuallyReloadTasks();
}

void GroupManagerPrivate::currentDesktopChanged(int newDesktop)
{
    //kDebug();
    if (!showOnlyCurrentDesktop) {
        return;
    }

    if (currentDesktop == newDesktop) {
        return;
    }

    if (!rootGroups[currentActivity].contains(newDesktop)) {
        kDebug() << "created new desk group";
        rootGroups[currentActivity][newDesktop] = new TaskGroup(q, "RootGroup", Qt::transparent);
        if (abstractSortingStrategy) {
            abstractSortingStrategy->handleGroup(rootGroups[currentActivity][newDesktop]);
        }
    }

    if (onlyGroupWhenFull) {
        QObject::disconnect(currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
        QObject::disconnect(currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
    }

    currentDesktop = newDesktop;

    foreach (LauncherItem *item, launchers) {
        if (item->shouldShow(q)) {
            rootGroups[currentActivity][currentDesktop]->add(item);
        } else {
            rootGroups[currentActivity][currentDesktop]->remove(item);
        }
    }

    if (onlyGroupWhenFull) {
        QObject::connect(currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
        QObject::connect(currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), q, SLOT(checkIfFull()));
    }
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << "Call reload";
#endif
    actuallyReloadTasks();
}


void GroupManagerPrivate::taskChanged(TaskPtr task, ::TaskManager::TaskChanges changes)
{
    //kDebug();
    bool takeAction = false;
    bool show = true;

#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << task->name() << (int)changes;
#endif
    
    if (showOnlyCurrentDesktop && changes & ::TaskManager::DesktopChanged) {
        takeAction = true;
        show = task->isOnCurrentDesktop();
        //kDebug() << task->visibleName() << "on" << TaskManager::self()->currentDesktop();
    }

    if (showOnlyCurrentActivity && changes & ::TaskManager::ActivitiesChanged) {
        takeAction = true;
        show = task->isOnCurrentActivity();
        //kDebug() << task->visibleName() << "on" << TaskManager::self()->currentDesktop();
    }

    if (showOnlyMinimized && changes & ::TaskManager::StateChanged) {
        //TODO: wouldn't it be nice to get notification of JUST minimization?
        takeAction = true;
        show = task->isMinimized();
    }

    if (showOnlyCurrentScreen && changes & ::TaskManager::GeometryChanged) {
        geometryTasks.insert(task.data());

        if (!screenTimer.isActive()) {
            screenTimer.start();
        }
    }

    if (changes & ::TaskManager::AttentionChanged) {
        // we show tasks anyway if they demand attention
        // so whenever our state changes ... try to re-adjust it
        takeAction = true;
        show = true;
    }

    // Some apps, eg. LibreOffice, change classClass/className after start-up...
    if (changes & ::TaskManager::ClassChanged) {
        // Instead of just moving  item (as what happend in the #if below), just remove and re-add the task.
        // This way the grouping happens properly.
        AbstractGroupableItem *item = currentRootGroup()->getMemberByWId(task->window());
        if(item && TaskItemType==item->itemType()) {
            item->resetLauncherCheck();
        }
        removeTask(task);
        addTask(task);
#if 0
        TaskItem *item = qobject_cast<TaskItem*>(currentRootGroup()->getMemberByWId(task->window()));
        if(item) {
            LauncherItem *prev=0L;
            item->resetTaskName();
            
            foreach (LauncherItem *launcher, launchers) {
                if(launcher->isAssociated(item)) {
                    prev=launcher;
                }
            }
            foreach (LauncherItem *launcher, launchers) {
                if(prev!=launcher && launcher->associateItemIfMatches(item)) {
                    if(prev) {
                        // Remove from previous launcher, as assigned to a new one...
                        prev->removeItemIfAssociated(item);
                        prev=0L;
                    }
                    // Place task at correct location...
                    if(!separateLaunchers && abstractSortingStrategy && GroupManager::ManualSorting==abstractSortingStrategy->type()) {
                        q->manualSortingRequest(item, launcherSort.indexOf(launcher->launcherUrl()));
                    }
                }
            }
        }
#endif
    }
    
    if (!takeAction) {
        return;
    }

    show = show && (!showOnlyCurrentScreen || task->isOnScreen(currentScreen));

    if (show) {
        //kDebug() << "add(task);";
        addTask(task);
    } else {
        //kDebug() << "remove(task);";
        removeTask(task);
    }
}

void GroupManager::setScreen(int screen)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << "Call reloadTasks";
#endif
    //kDebug() << "new Screen: " << screen;
    d->currentScreen = screen;
    d->reloadTasks();
}

int GroupManager::screen() const
{
    return d->currentScreen;
}

void GroupManagerPrivate::checkScreenChange()
{
    //kDebug();
    foreach (Task *task, geometryTasks) {
        if (task->isOnScreen(currentScreen)) {
            addTask(TaskPtr(task));
        } else {
            removeTask(TaskPtr(task));
        }
    }
}

void GroupManager::reconnect()
{
    //kDebug();
    disconnect(TaskManager::self(), SIGNAL(desktopChanged(int)),
               this, SLOT(currentDesktopChanged(int)));
    disconnect(TaskManager::self(), SIGNAL(activityChanged(QString)),
               this, SLOT(currentActivityChanged(QString)));
    disconnect(TaskManager::self(), SIGNAL(windowChanged(TaskPtr,::TaskManager::TaskChanges)),
               this, SLOT(taskChanged(TaskPtr,::TaskManager::TaskChanges)));

    if (d->showOnlyCurrentDesktop || d->showOnlyMinimized || d->showOnlyCurrentScreen || d->showOnlyCurrentActivity) {
        // listen to the relevant task manager signals
        if (d->showOnlyCurrentDesktop) {
            connect(TaskManager::self(), SIGNAL(desktopChanged(int)),
                    this, SLOT(currentDesktopChanged(int)));
        }
        if (d->showOnlyCurrentActivity) {
            connect(TaskManager::self(), SIGNAL(activityChanged(QString)),
                    this, SLOT(currentActivityChanged(QString)));
        }

        connect(TaskManager::self(), SIGNAL(windowChanged(TaskPtr,::TaskManager::TaskChanges)),
                this, SLOT(taskChanged(TaskPtr,::TaskManager::TaskChanges)));
    }

    TaskManager::self()->setTrackGeometry(d->showOnlyCurrentScreen, d->configToken);

    if (!d->showOnlyCurrentScreen) {
        d->geometryTasks.clear();
    }
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << "Call reloadTasks";
#endif
    d->reloadTasks();
}

KConfigGroup GroupManager::config() const
{
    return KConfigGroup();
}

bool GroupManager::addLauncher(const KUrl &url, const QString &key, int index)
{
    if (url.isEmpty() || launchersLocked()) {
        return false;
    }

    if(!QFile::exists(url.toLocalFile())) {
        return false;
    }

    if(!d->launcherSort.contains(url)) {
        if(index>=0 && index<d->launcherSort.count()) {
            d->launcherSort.insert(index, Launcher(url, key));
        } else {
            d->launcherSort.append(Launcher(url, key));
        }
        d->saveLauncherConfig();
    }

    LauncherItem *launcher = d->launchers.value(url); // Do not insert launchers twice
    if (!launcher) {
        launcher = new LauncherItem(d->currentRootGroup(), url);

        if (!launcher->isValid()) {
            delete launcher;
            return false;
        }

        QStack<TaskGroup *> groups;
        groups.push(d->currentRootGroup());
        while (!groups.isEmpty()) {
            TaskGroup *group = groups.pop();

            foreach (AbstractGroupableItem *item, group->members()) {
                if (item->itemType() == GroupItemType) {
                    groups.push(static_cast<TaskGroup *>(item));
                } else {
                    launcher->associateItemIfMatches(item);
                }
            }
        }

        d->launchers.insert(url, launcher);
        connect(launcher, SIGNAL(associationChanged()), this, SLOT(launcherVisibilityChange()));
        d->checkLauncherVisibility(launcher);

        if(!d->separateLaunchers && d->abstractSortingStrategy && ManualSorting==d->abstractSortingStrategy->type()) {
            // Ensure item is placed where launcher would be...
            foreach (AbstractGroupableItem *item, d->rootGroups[d->currentActivity][d->currentDesktop]->members()) {
                if(LauncherItemType!=item->itemType() && item->launcherUrl()==url) {
                    manualSortingRequest(item, d->launcherSort.indexOf(url));
                    break;
                }
            }
            
            if(!d->readingLauncherConfig) {
                emit launchersChanged();
            }
        }
    }

    return launcher;
}

void GroupManager::removeLauncher(const KUrl &url)
{
    if(launchersLocked()) {
        return;
    }

    LauncherItem *launcher = d->launchers.value(url);
    if (!launcher) {
        return;
    }

    if(d->launcherSort.contains(url)) {
        d->launcherSort.removeAll(url);
        d->saveLauncherConfig();
    }

    d->launchers.remove(url);

    typedef QHash<int,TaskGroup*> Metagroup;
    foreach (Metagroup metagroup, d->rootGroups) {
        foreach (TaskGroup *rootGroup, metagroup) {
            rootGroup->remove(launcher);
        }
    }

    launcher->deleteLater();
    
    if(!d->separateLaunchers && d->abstractSortingStrategy && ManualSorting==d->abstractSortingStrategy->type()) {
        // Ensure item is placed at end of launchers...
        foreach (AbstractGroupableItem *item, d->rootGroups[d->currentActivity][d->currentDesktop]->members()) {
            if(LauncherItemType!=item->itemType() && item->launcherUrl()==url) {
                manualSortingRequest(item, d->launcherSort.count());
                break;
            }
        }

        if(!d->readingLauncherConfig) {
            emit launchersChanged();
        }
    }
}

void GroupManagerPrivate::sycocaChanged(const QStringList &types)
{
    if(types.contains("apps")) {
        KUrl::List removals;
        foreach (LauncherItem *launcher, launchers) {
            if(!QFile::exists(launcher->launcherUrl().toLocalFile())) {
                removals << launcher->launcherUrl();
            }
        }

        foreach (const KUrl &url, removals) {
            q->removeLauncher(url);
        }
    }
}

void GroupManagerPrivate::launcherVisibilityChange()
{
    checkLauncherVisibility(qobject_cast<LauncherItem *>(q->sender()));
}

void GroupManagerPrivate::checkLauncherVisibility(LauncherItem *launcher)
{
    if (!launcher) {
        return;
    }

    if (launcher->shouldShow(q)) {
        rootGroups[currentActivity][currentDesktop]->add(launcher);
    } else {
        rootGroups[currentActivity][currentDesktop]->remove(launcher);
    }
}

bool GroupManager::launcherExists(const KUrl &url) const
{
    return d->launchers.value(url);
}

void GroupManager::readLauncherConfig(const KConfigGroup &cg)
{
    KConfigGroup conf = d->launcherConfig(cg);
    if (!conf.isValid()) {
        return;
    }

    QList<Launcher> old=d->launcherSort;
    d->readLauncherConfig(conf);
    bool launcherOrderChanged=old!=d->launcherSort;

    if(launcherOrderChanged) {
        // Launcher order has changed, so we need to re-load tasks - so that these are sorted properly.
        // First clear current group...
        #ifdef ICON_ENABLE_DBUS_DBUG
        ICN_LIB_DBUG << "clear current group";
        #endif
        d->currentRootGroup()->clear();
    }
    // prevents re-writing the results out
    d->readingLauncherConfig = true;
    foreach (const Launcher &url, d->launcherSort) {
        addLauncher(url);
    }

    //d->readingLauncherConfig = false;

    // a bit paranoiac, perhaps, but we check the removals first and then
    // remove the launchers after that scan because Qt's iterators operate
    // on a copy of the list and/or don't like multiple iterators messing
    // with the same list. we might get away with just calling removeLauncher
    // immediately without the removals KUrl::List, but this is known safe
    // and not a performance bottleneck
    KUrl::List removals;
    foreach (LauncherItem *launcher, d->launchers) {
        if (!d->launcherSort.contains(launcher->launcherUrl())) {
            removals << launcher->launcherUrl();
        }
    }

    foreach (const KUrl &url, removals) {
        removeLauncher(url);
    }
    
    d->readingLauncherConfig = false;
    
    if(launcherOrderChanged) {
        #ifdef ICON_ENABLE_DBUS_DBUG
        ICN_LIB_DBUG << "reload tasks";
        #endif
        // ...now reload all tasks, as they were cleared...
        d->actuallyReloadTasks();
    }
}

int GroupManager::launcherIndex(const KUrl &url) const
{
    return d->launcherSort.indexOf(url);
}

int GroupManager::launcherCount() const
{
    return d->launcherSort.count();
}

bool GroupManager::launchersLocked() const
{
    return d->launchersLocked && !d->readingLauncherConfig;
}

void GroupManager::setLaunchersLocked(bool l)
{
    if(l!=d->launchersLocked) {
        d->launchersLocked=l;
        d->saveLauncherConfig();
    }
}

const QList<Launcher> & GroupManager::launchers() const
{
    return d->launcherSort;
}

KUrl GroupManager::launcherForKey(const QString &key) const
{
    foreach(Launcher l, d->launcherSort) {
        if(l.key==key) {
            return l;
        }
    }
    
    return KUrl();
}

bool GroupManager::itemIsAssociatedwithLauncher(AbstractGroupableItem *item)
{
    if(item) {
        switch(item->itemType()) {
            case LauncherItemType:
                return true;
            case GroupItemType: {
                foreach(AbstractGroupableItem *i, static_cast<TaskGroup *>(item)->members()) {
                    if(itemIsAssociatedwithLauncher(i)) {
                        return true;
                    }
                }
                break;
            }
            case TaskItemType: {
                foreach (LauncherItem *launcher, d->launchers) {
                    if(launcher->isAssociated(item)) {
                        return true;
                    }
                }
            }
        }
    }
    
    return false;
}

void GroupManager::moveLauncher(const KUrl &url, int newIndex)
{
    if(!url.isValid()) {
        return;
    }
    int oldIndex=d->launcherSort.indexOf(url);
    
    if(oldIndex>=0 && newIndex!=oldIndex) {
        d->launcherSort.move(oldIndex, newIndex);
        d->saveLauncherConfig();
    }
}

bool GroupManager::separateLaunchers() const
{
    return d->separateLaunchers;
}

void GroupManager::setSeparateLaunchers(bool s)
{
    d->separateLaunchers=s;
}

bool GroupManager::groupAttention() const
{
    return d->groupAttention;
}

void GroupManager::setGroupAttention(bool s)
{
    d->groupAttention=s;
}

void GroupManager::createConfigurationInterface(KConfigDialog *parent)
{
    new LauncherConfig(parent);
}

KConfigGroup GroupManagerPrivate::launcherConfig(const KConfigGroup &config)
{
    KConfigGroup cg = config.isValid() ? config : q->config();
    if (!cg.isValid()) {
        return cg;
    }

    KConfigGroup lg(&cg, "Launchers");

    // Move any old (pre 0.4.0 config (which still retained KDE-taskmanager config items)
    // to new format.
    if(lg.isValid() && lg.hasGroup("Config")) {
        KConfigGroup cfg(&lg, "Config");
        
        if(cfg.isValid() && cfg.hasKey("Order")) {
            QStringList order=cfg.readEntry("Order", QStringList());
            bool        locked=cfg.readEntry("LockLaunchers", false);
            lg.writeEntry("Items", order);
            lg.writeEntry("Locked", locked);
            cfg.deleteGroup();
            
            foreach (const QString &key, lg.keyList()) {
                if(key!="Items" && key!="Locked") {
                    lg.deleteEntry(key);
                }
            }
            cg.sync();
        }
    }

    return lg;
}

void GroupManagerPrivate::saveLauncherConfig()
{
    if (readingLauncherConfig) {
        return;
    }

    KConfigGroup cg = launcherConfig();
    if (!cg.isValid()) {
        return;
    }
    saveLauncherConfig(cg);
}

void GroupManagerPrivate::saveLauncherConfig(KConfigGroup &cg)
{
    QStringList launchers;
    QStringList keys;
    foreach (Launcher url, launcherSort) {
        QString u(url.url());
        if(!u.isEmpty()) {
            if(!url.key.isEmpty()) {
                launchers.append('{'+url.key+'}');
            }
            launchers.append(u);
        }
    }
    
    cg.writeEntry("Items", launchers);
    cg.writeEntry("Locked", launchers.count() && launchersLocked);

#ifdef ICON_ENABLE_DBUS_DBUG
    if(GroupManager::debug()) {
        int i=0;
        foreach(Launcher u, launcherSort) {
            ICN_LIB_STD_DBUG << '(' << i << ')' << u.url() << '{' << u.key << '}';
        }
    }
#endif
    emit q->configChanged();
}

void GroupManagerPrivate::readLauncherConfig()
{
    if (readingLauncherConfig) {
        return;
    }

    KConfigGroup cg = launcherConfig();
    if (!cg.isValid()) {
        return;
    }
    
    readLauncherConfig(cg);
}
    
void GroupManagerPrivate::readLauncherConfig(KConfigGroup &cg)
{
    QStringList lo=cg.readEntry("Items", QStringList());
    bool        altered=false;
    QString     key;
    launcherSort.clear();
    foreach (QString o, lo) {
        if(!o.isEmpty()) {
            if(o.startsWith('{') && o.endsWith('}')) {
                key=o.mid(1, o.length()-2);
                continue;
            }
            // Attempt to fix pre 0.5.0 IconTasks matching of amarok_containers.desktop before amarok.desktop
            // If a user has this, see if amarok.desktop exists, and if so use that...
            // This codeblock should be removed at a later date (v 0.7.0 ??)
            if(o.endsWith("amarok_containers.desktop")) {
                QString fix(o);
                fix.replace("amarok_containers.desktop", "amarok.desktop");
                KUrl url(fix);
                if(QFile::exists(url.toLocalFile())) {
                    launcherSort.append(Launcher(url, key));
                    altered=true;
                    continue;
                }
            }

            if(!key.isEmpty()) {
                // Fix <0.7.2 error where brackets were not removed when reading in!!!!
                key.replace("{", "");
                key.replace("}", "");
                // We have stored the 'wmClass' for the app, see if the user set a manual override, if so use that!
                KConfig cfg("taskmanagerrc");
                KConfigGroup grp(&cfg, "Mapping");
                QString mapped=grp.readEntry(key, QString());
        
                if(mapped.endsWith(".desktop") && mapped!=o) {
                    o=mapped;
                    altered=true;
                }
            }

            launcherSort.append(Launcher(KUrl(o), key));
            key=QString();
        }
    }
    launchersLocked=launcherSort.count() && cg.readEntry("Locked", false);
#ifdef ICON_ENABLE_DBUS_DBUG
    if(GroupManager::debug()) {
        int i=0;
        foreach(Launcher u, launcherSort) {
            ICN_LIB_STD_DBUG << '(' << i++ << ')' << u.url() << '{' << u.key << '}';
        }
    }
#endif

    if(altered) {
        saveLauncherConfig(cg);
    }
}

bool GroupManager::onlyGroupWhenFull() const
{
    return d->onlyGroupWhenFull;
}

void GroupManager::setOnlyGroupWhenFull(bool onlyGroupWhenFull)
{
    //kDebug() << onlyGroupWhenFull;
    if (d->onlyGroupWhenFull == onlyGroupWhenFull) {
        return;
    }

    d->onlyGroupWhenFull = onlyGroupWhenFull;

    disconnect(d->currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), this, SLOT(checkIfFull()));
    disconnect(d->currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), this, SLOT(checkIfFull()));

    if (onlyGroupWhenFull) {
        connect(d->currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), this, SLOT(checkIfFull()));
        connect(d->currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), this, SLOT(checkIfFull()));
        d->checkIfFull();
    }
}

void GroupManager::setFullLimit(int limit)
{
    //kDebug() << limit;
    d->groupIsFullLimit = limit;
    if (d->onlyGroupWhenFull) {
        d->checkIfFull();
    }
}

void GroupManagerPrivate::checkIfFull()
{
    // Start a timer so that if we have been triggered by a layouting
    // we give time for it to finish up instead of starting a new one
    // right away.
    checkIfFullTimer.start();
}

void GroupManagerPrivate::actuallyCheckIfFull()
{
    //kDebug();
    if (!onlyGroupWhenFull ||
        groupingStrategy != GroupManager::ProgramGrouping ||
        changingGroupingStrategy) {
        return;
    }

    if (currentRootGroup()->totalSize() >= groupIsFullLimit) {
        if (!abstractGroupingStrategy) {
            geometryTasks.clear();
            q->setGroupingStrategy(GroupManager::ProgramGrouping);
        }
    } else if (abstractGroupingStrategy) {
        geometryTasks.clear();
        q->setGroupingStrategy(GroupManager::NoGrouping);
        //let the visualization think we still use the programGrouping
        groupingStrategy = GroupManager::ProgramGrouping;
    }
}

bool GroupManager::showOnlyCurrentScreen() const
{
    return d->showOnlyCurrentScreen;
}

void GroupManager::setShowOnlyCurrentScreen(bool showOnlyCurrentScreen)
{
    d->showOnlyCurrentScreen = showOnlyCurrentScreen;
    reconnect();
}

bool GroupManager::showOnlyCurrentDesktop() const
{
    return d->showOnlyCurrentDesktop;
}

void GroupManager::setShowOnlyCurrentDesktop(bool showOnlyCurrentDesktop)
{
    d->showOnlyCurrentDesktop = showOnlyCurrentDesktop;
    reconnect();
}

bool GroupManager::showOnlyCurrentActivity() const
{
    return d->showOnlyCurrentActivity;
}

void GroupManager::setShowOnlyCurrentActivity(bool showOnlyCurrentActivity)
{
    d->showOnlyCurrentActivity = showOnlyCurrentActivity;
    reconnect();
}

bool GroupManager::showOnlyMinimized() const
{
    return d->showOnlyMinimized;
}

void GroupManager::setShowOnlyMinimized(bool showOnlyMinimized)
{
    d->showOnlyMinimized = showOnlyMinimized;
    reconnect();
}

GroupManager::TaskSortingStrategy GroupManager::sortingStrategy() const
{
    return d->sortingStrategy;
}

AbstractSortingStrategy* GroupManager::taskSorter() const
{
    return d->abstractSortingStrategy;
}

void GroupManager::setSortingStrategy(TaskSortingStrategy sortOrder)
{
    //kDebug() << sortOrder;

    if (d->abstractSortingStrategy) {
        if (d->abstractSortingStrategy->type() == sortOrder) {
            return;
        }

        d->abstractSortingStrategy->deleteLater();
        d->abstractSortingStrategy = 0;
    }

    switch (sortOrder) {
        case ManualSorting:
            d->abstractSortingStrategy = new ManualSortingStrategy(this);
            break;

        case AlphaSorting:
            d->abstractSortingStrategy = new AlphaSortingStrategy(this);
            break;

        case DesktopSorting:
            d->abstractSortingStrategy = new DesktopSortingStrategy(this);
            break;

        case NoSorting: //manual and no grouping result both in non automatic grouping
            break;

        default:
            kDebug() << "Invalid Strategy";
    }
    if (d->abstractSortingStrategy) {
        typedef QHash<int,TaskGroup*> Metagroup;
        foreach (Metagroup metagroup, d->rootGroups) {
            foreach (TaskGroup *group, metagroup) {
                d->abstractSortingStrategy->handleGroup(group);
            }
        }
    }

    d->sortingStrategy = sortOrder;
    reconnect();
}

GroupManager::TaskGroupingStrategy GroupManager::groupingStrategy() const
{
    return d->groupingStrategy;
}

AbstractGroupingStrategy* GroupManager::taskGrouper() const
{
    return d->abstractGroupingStrategy;
}


void GroupManager::setGroupingStrategy(TaskGroupingStrategy strategy)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_DBUG << strategy;
#endif
    if (d->changingGroupingStrategy ||
        (d->abstractGroupingStrategy && d->abstractGroupingStrategy->type() == strategy)) {
        return;
    }

    d->changingGroupingStrategy = true;

    //kDebug() << strategy << kBacktrace();
    if (d->onlyGroupWhenFull) {
        disconnect(d->currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), this, SLOT(checkIfFull()));
        disconnect(d->currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), this, SLOT(checkIfFull()));
    }

    if (d->abstractGroupingStrategy) {
        disconnect(d->abstractGroupingStrategy, 0, this, 0);
        d->abstractGroupingStrategy->destroy();
        d->abstractGroupingStrategy = 0;
    }

    switch (strategy) {
        case ManualGrouping:
            d->abstractGroupingStrategy = new ManualGroupingStrategy(this);
            break;

        case ProgramGrouping:
            d->abstractGroupingStrategy = new ProgramGroupingStrategy(this);
            break;

        case KustodianGrouping:
            d->abstractGroupingStrategy = new KustodianGroupingStrategy(this);
            break;

        case NoGrouping:
            break;

        default:
            kDebug() << "Strategy not implemented";
    }

    d->groupingStrategy = strategy;
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_STD_DBUG << "Call reload";
#endif
    d->actuallyReloadTasks();

    if (d->onlyGroupWhenFull) {
        connect(d->currentRootGroup(), SIGNAL(itemAdded(AbstractGroupableItem*)), this, SLOT(checkIfFull()));
        connect(d->currentRootGroup(), SIGNAL(itemRemoved(AbstractGroupableItem*)), this, SLOT(checkIfFull()));
    }

    d->changingGroupingStrategy = false;
}

#ifdef ICON_ENABLE_DBUS_DBUG
void GroupManager::dump(QTextStream &str) const
{
    str << "<" << QString(metaObject()->className()).replace("::", "__") << ">";
    QHash<KUrl, LauncherItem *>::ConstIterator lit(d->launchers.constBegin()),
                                               lend(d->launchers.constEnd());
    str << "<launchers>";
    for(; lit!=lend; ++lit) {
        lit.value()->dump(str);
    }
    str << "</launchers>";
    QSet<Task *>::ConstIterator git(d->geometryTasks.constBegin()),
                                gend(d->geometryTasks.constEnd());
    str << "<tasks>";
    for(; git!=gend; ++git) {
        str << "<task window=\"" << (*git)->window()
            << "\" ptr=\"" << (void *)(*git)
            << "\" visibleName=\"" << Qt::escape((*git)->visibleName())
            << "\" name=\"" << Qt::escape((*git)->name())
            << "\" className=\"" << Qt::escape((*git)->className())
            << "\" classClass=\"" << Qt::escape((*git)->classClass())
            << "\" pid=\"" << (*git)->pid()
            << "\" isMaximized=\"" << (*git)->isMaximized()
            << "\" isMinimized=\"" << (*git)->isMinimized()
            << "\" isIconified=\"" << (*git)->isIconified()
            << "\" isActive=\"" << (*git)->isActive()
            << "\" isOnCurrentDesktop=\"" << (*git)->isOnCurrentDesktop()
            << "\" isOnAllDesktops=\"" << (*git)->isOnAllDesktops()
            << "\" desktop=\"" << (*git)->desktop()
            << "\" demandsAttention=\"" << (*git)->demandsAttention()
            << "\" showInTaskbar=\"" << (*git)->showInTaskbar()
            << "\" isOnCurrentActivity=\"" << (*git)->isOnCurrentActivity()
            << "\" isOnAllActivities=\"" << (*git)->isOnAllActivities()
            << "\" />";
    }
    str << "</tasks>";
    QHash<QString, QHash<int, TaskGroup*> >::ConstIterator ait(d->rootGroups.constBegin()),
                                                           aend(d->rootGroups.constEnd());
                                                           
    for(; ait!=aend; ++ait) {
        QString activity=ait.key();
        QHash<int, TaskGroup*>::ConstIterator dit(ait.value().constBegin()),
                                              dend(ait.value().constEnd());
        for(; dit!=dend; ++dit) {
            int desktop=dit.key();
            str << "<rootGroup activity=\"" << activity << "\" desktop=\"" << desktop << "\" ";
            if(desktop==d->currentDesktop && activity==d->currentActivity) {
                str << "current=\"true\"";
            }
            str << ">\n";
            dit.value()->dump(str);
            str << "</rootGroup>";
        }
    }
    
    str << "</" << QString(metaObject()->className()).replace("::", "__") << ">";
}

int GroupManager::m_debug=0;
#endif
    
} // TaskManager namespace

#include "groupmanager.moc"
