/*****************************************************************

Copyright 2008 Christian Mollekopf <chrigi_1@hotmail.com>
Copyright (C) 2011 Craig Drummond <craig.p.drummond@gmail.com> 

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

// Own
#include "taskitem.h"

#include <KConfig>
#include <KConfigGroup>
#include <KDebug>
#include <KDesktopFile>
#include <KService>
#include <KServiceTypeTrader>
#include <KStandardDirs>

#include <QtCore/QFile>
#include <QtCore/QTimer>

#include "config.h"
#include "groupmanager.h"
#ifdef ICON_ENABLE_DBUS_DBUG
#include <QTextDocument>
#include "taskgroup.h"
#endif

namespace TaskManager
{


class TaskItem::Private
{
public:
    Private()
        : startupTask(0)
        , checkedForLauncher(false)
    {
    }

    QWeakPointer<Task> task;
    StartupPtr startupTask;
    KUrl launcherUrl;
    bool checkedForLauncher;
    QString taskName;
};


TaskItem::TaskItem(QObject *parent, TaskPtr task)
    : AbstractGroupableItem(parent),
      d(new Private)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_DBUG << "Std";
#endif
    setTaskPointer(task);
}


TaskItem::TaskItem(QObject *parent, StartupPtr task)
    : AbstractGroupableItem(parent),
      d(new Private)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_DBUG << "StartUp";
#endif
    d->startupTask = task;
    connect(task.data(), SIGNAL(changed(::TaskManager::TaskChanges)), this, SIGNAL(changed(::TaskManager::TaskChanges)));
    connect(task.data(), SIGNAL(destroyed(QObject*)), this, SLOT(taskDestroyed()), Qt::QueuedConnection); //this item isn't useful anymore if the Task was closed
}

TaskItem::~TaskItem()
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_DBUG;
#endif

    emit destroyed(this);
    //kDebug();
  /*  if (parentGroup()){
        parentGroup()->remove(this);
    }*/
    delete d;
}

void TaskItem::taskDestroyed()
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_DBUG;
#endif
    d->startupTask = 0;
    d->task.clear();
//     deleteLater();
    QTimer::singleShot(0, this, SLOT(deleteLater()));
}

void TaskItem::setTaskPointer(TaskPtr task)
{
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_DBUG << (void *)task.data();
#endif
    if (d->startupTask) {
        disconnect(d->startupTask.data(), 0, this, 0);
        d->startupTask = 0;
    }

    if (d->task.data() != task.data()) {
        if (d->task) {
            disconnect(d->task.data(), 0, this, 0);
        }

        d->task = task.data();

        if (task) {
            connect(task.data(), SIGNAL(changed(::TaskManager::TaskChanges)), this, SIGNAL(changed(::TaskManager::TaskChanges)));
            connect(task.data(), SIGNAL(destroyed(QObject*)), this, SLOT(taskDestroyed()), Qt::QueuedConnection);
            emit gotTaskPointer();
        }
    }

    if (!d->task) {
#ifdef ICON_ENABLE_DBUS_DBUG
        ICN_LIB_DBUG << "No task, deleteLater";
#endif
//         deleteLater();
        QTimer::singleShot(0, this, SLOT(deleteLater()));
    }
}

TaskPtr TaskItem::task() const
{
    if (!d->task) {
        return TaskPtr(0);
    }
    return TaskPtr(d->task.data());
}

StartupPtr TaskItem::startup() const
{
    /*
    if (d->startupTask.isNull()) {
        kDebug() << "pointer is Null";
    }
    */
    return d->startupTask;
}

bool TaskItem::isStartupItem() const
{
    return !d->startupTask.isNull();
}

WindowList TaskItem::winIds() const
{
    if (!d->task) {
        kDebug() << "no winId: probably startup task";
        return WindowList();
    }
    WindowList list;
    list << d->task.data()->window();
    return list;
}

QIcon TaskItem::icon() const
{
    if (d->task) {
        return d->task.data()->icon();
    }

    if (d->startupTask) {
        return d->startupTask->icon();
    }

    return QIcon();
}

QString TaskItem::name() const
{
    if (d->task) {
        return d->task.data()->visibleName();
    }

    if (d->startupTask) {
        return d->startupTask->text();
    }

    return QString();
}

QString TaskItem::taskName() const
{
    if (d->taskName.isEmpty()) {
        KUrl lUrl=launcherUrl();

        if (!lUrl.isEmpty() && lUrl.isLocalFile() && KDesktopFile::isDesktopFile(lUrl.toLocalFile())) {
            KDesktopFile f(lUrl.toLocalFile());

            if (f.tryExec()) {
                d->taskName = f.readName();
            }
        }
        if(d->taskName.isEmpty() && d->task) {
            d->taskName = d->task.data()->classClass().toLower();
        }
    }
    
    return d->taskName;
}

void TaskItem::resetTaskName()
{
    d->taskName=QString();
}

ItemType TaskItem::itemType() const
{
    return TaskItemType;
}

bool TaskItem::isGroupItem() const
{
    return false;
}

void TaskItem::setShaded(bool state)
{
    if (!d->task) {
        return;
    }
    d->task.data()->setShaded(state);
}

void TaskItem::toggleShaded()
{
    if (!d->task) {
        return;
    }
    d->task.data()->toggleShaded();
}

bool TaskItem::isShaded() const
{
    if (!d->task) {
        return false;
    }
    return d->task.data()->isShaded();
}

void TaskItem::toDesktop(int desk)
{
    if (!d->task) {
        return;
    }
    d->task.data()->toDesktop(desk);
}

bool TaskItem::isOnCurrentDesktop() const
{
    return d->task && d->task.data()->isOnCurrentDesktop();
}

bool TaskItem::isOnAllDesktops() const
{
    return d->task && d->task.data()->isOnAllDesktops();
}

int TaskItem::desktop() const
{
    if (!d->task) {
        return 0;
    }
    return d->task.data()->desktop();
}

void TaskItem::setMaximized(bool state)
{
    if (!d->task) {
        return;
    }
    d->task.data()->setMaximized(state);
}

void TaskItem::toggleMaximized()
{
    if (!d->task) {
        return;
    }
    d->task.data()->toggleMaximized();
}

bool TaskItem::isMaximized() const
{
    return d->task && d->task.data()->isMaximized();
}

void TaskItem::setMinimized(bool state)
{
    if (!d->task) {
        return;
    }
    d->task.data()->setIconified(state);
}

void TaskItem::toggleMinimized()
{
    if (!d->task) {
        return;
    }
    d->task.data()->toggleIconified();
}

bool TaskItem::isMinimized() const
{
    if (!d->task) {
        return false;
    }
    return d->task.data()->isMinimized();
}

void TaskItem::setFullScreen(bool state)
{
    if (!d->task) {
        return;
    }
    d->task.data()->setFullScreen(state);
}

void TaskItem::toggleFullScreen()
{
    if (!d->task) {
        return;
    }
    d->task.data()->toggleFullScreen();
}

bool TaskItem::isFullScreen() const
{
    if (!d->task) {
        return false;
    }
    return d->task.data()->isFullScreen();
}

void TaskItem::setKeptBelowOthers(bool state)
{
    if (!d->task) {
        return;
    }
    d->task.data()->setKeptBelowOthers(state);
}

void TaskItem::toggleKeptBelowOthers()
{
    if (!d->task) {
        return;
    }
    d->task.data()->toggleKeptBelowOthers();
}

bool TaskItem::isKeptBelowOthers() const
{
    if (!d->task) {
        return false;
    }
    return d->task.data()->isKeptBelowOthers();
}

void TaskItem::setAlwaysOnTop(bool state)
{
    if (!d->task) {
        return;
    }
    d->task.data()->setAlwaysOnTop(state);
}

void TaskItem::toggleAlwaysOnTop()
{
    if (!d->task) {
        return;
    }
    d->task.data()->toggleAlwaysOnTop();
}

bool TaskItem::isAlwaysOnTop() const
{
    if (!d->task) {
        return false;
    }
    return d->task.data()->isAlwaysOnTop();
}

bool TaskItem::isActionSupported(NET::Action action) const
{
    return d->task && d->task.data()->info().actionSupported(action);
}

void TaskItem::addMimeData(QMimeData *mimeData) const
{
    if (!d->task) {
        return;
    }

    d->task.data()->addMimeData(mimeData);
}

static KService::List getServicesViaPid(int pid)
{
    // Attempt to find using commandline...
    QFile f(QString("/proc/%1/cmdline").arg(pid));
    KService::List services;
    
    if(f.open(QIODevice::ReadOnly)) {
        QByteArray bytes=f.read(1024);
        int        firstSpace=0;
        
        if(bytes.length()>2) {
            for(int i=0; i<bytes.length()-1; ++i) {
                if(0==bytes[i]) {
                    if(0==firstSpace) {
                        firstSpace=i;
                    }
                    bytes[i]=' ';
                }
            }
            
            QString cmdline(bytes);
            
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ Exec)").arg(cmdline));
#ifdef ICON_ENABLE_DBUS_DBUG
            if(GroupManager::launcherDebug()) qDebug() << __PRETTY_FUNCTION__ << "Query Exec via cmdline" << cmdline << (int)(services.empty() ? 0 : 1);
#endif
            if(services.empty()) {
                // Could not find with complete commandline, so strip out path part...
                int slash=cmdline.lastIndexOf('/', firstSpace);
                if(slash>0) {
                    services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ Exec)").arg(cmdline.mid(slash+1)));
#ifdef ICON_ENABLE_DBUS_DBUG
                    if(GroupManager::launcherDebug()) qDebug() << __PRETTY_FUNCTION__ << "Query Exec via cmdline(no path)" << cmdline.mid(slash+1) << (int)(services.empty() ? 0 : 1);
#endif
                }
            }
            
            if(services.empty() && firstSpace) {
                // Could not find with arguments, so try without...
                cmdline=cmdline.left(firstSpace);
                services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ Exec)").arg(cmdline));
#ifdef ICON_ENABLE_DBUS_DBUG
                if(GroupManager::launcherDebug()) qDebug() << __PRETTY_FUNCTION__ << "Query Exec via cmdline(no args)" << cmdline << (int)(services.empty() ? 0 : 1);
#endif
                int slash=cmdline.lastIndexOf('/');
                if(slash>0) {
                    services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ Exec)").arg(cmdline.mid(slash+1)));
#ifdef ICON_ENABLE_DBUS_DBUG
                    if(GroupManager::launcherDebug()) qDebug() << __PRETTY_FUNCTION__ << "Query Exec via cmdline(no args and no path)" << cmdline.mid(slash+1) << (int)(services.empty() ? 0 : 1);
#endif
                }
            }
        }
    }
    
    return services;
}

void TaskItem::setLauncherUrl(const KUrl &url, bool saveMapping) 
{
    if (!d->launcherUrl.isEmpty()) {
        return;
    }
    d->launcherUrl=url;
    d->taskName=QString(); // Cause name to be re-generated...

    if(saveMapping) {
        KConfig cfg("taskmanagerrc");
        KConfigGroup grp(&cfg, "Mapping");
        grp.writeEntry(d->task.data()->classClass()+"::"+d->task.data()->className(), url.url());
        cfg.sync();
    }
}

KUrl TaskItem::launcherUrl() const
{
    if (!d->task && !isStartupItem()) {
        return KUrl();
    }

    if (!d->launcherUrl.isEmpty() || d->checkedForLauncher) {
        return d->launcherUrl;
    }
#ifdef ICON_ENABLE_DBUS_DBUG
    if(d->task) {
        ICN_LIB_L_DBUG << "Get launcher URL for" << d->task.data()->classClass() << d->task.data()->className();
    } else {
        ICN_LIB_L_DBUG << "Get launcher URL for (startup)" << name();
    }
#endif

    // Search for applications which are executable and case-insensitively match the windowclass of the task and
    // See http://techbase.kde.org/Development/Tutorials/Services/Traders#The_KTrader_Query_Language
    KService::List services;
    bool triedPid=false;

    // Set a flag so that we remeber that we have already checked for a launcher. This is becasue if we fail, then
    // we will keep on failing so the isEmpty()  check above is not enough.
    d->checkedForLauncher=true;

    if(d->task && !d->task.data()->classClass().isEmpty()) {
        // Check to see if this wmClass matched a saved one...      
        KConfig cfg("taskmanagerrc");
        KConfigGroup grp(&cfg, "Mapping");
        KConfigGroup set(&cfg, "Settings");

        // Some apps have different laucnhers depending upon comandline...
        QStringList matchCommandLineFirst=set.readEntry("MatchCommandLineFirst", QStringList());
        if(matchCommandLineFirst.contains(d->task.data()->classClass())) {
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Matching on commandline first";
#endif
            triedPid=true;
            services=getServicesViaPid(d->task.data()->pid());
        }

        // If the user has manualy set a mapping, respect this first...
        QString mapped(grp.readEntry(d->task.data()->classClass()+"::"+d->task.data()->className(), QString()));
        
        if(mapped.endsWith(".desktop")) {
            d->launcherUrl=mapped;
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Matched on full " << d->launcherUrl;
#endif
            return d->launcherUrl;
        }
        
        if(mapped.isEmpty()) {
            mapped=grp.readEntry(d->task.data()->classClass(), QString());
        
            if(mapped.endsWith(".desktop")) {
                d->launcherUrl=mapped;
#ifdef ICON_ENABLE_DBUS_DBUG
                ICN_LIB_L_DBUG << "Matched on class " << d->launcherUrl;
#endif
                return d->launcherUrl;
            }
        }
        
        // Some apps, such as Wine, cannot use className to map to launcher name - as Wine itself is not a GUI app
        // So, Settings/ManualOnly lists window classes where the user will always ahve to manual set the launcher...
        QStringList manualOnly=set.readEntry("ManualOnly", QStringList());
        
        if(manualOnly.contains(d->task.data()->classClass())) {
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Need to set manually, controlled via ManualOnly";
#endif
            return d->launcherUrl;
        }
        
        if(!mapped.isEmpty()) {
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ DesktopEntryName)").arg(mapped));
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Query DesktopEntryName via mapped" << mapped << (int)(services.empty() ? 0 : 1);
#endif
        }

        if(!mapped.isEmpty() && services.empty()) {
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ Name)").arg(mapped));
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Query Name via mapped" << mapped << (int)(services.empty() ? 0 : 1);
#endif
        }

        if(services.empty() && qobject_cast<GroupManager *>(parent())) {
            KUrl savedUrl=static_cast<GroupManager *>(parent())->launcherForKey(d->task.data()->classClass());
            if(savedUrl.isValid()) {
                d->launcherUrl=savedUrl;
#ifdef ICON_ENABLE_DBUS_DBUG
                ICN_LIB_L_DBUG << "Matched to wmClass in list of launchers " << d->launcherUrl;
#endif
                return d->launcherUrl;
            }
        }

        // To match other docks (docky, unity, etc.) attempt to match on DesktopEntryName first...
        if(services.empty()) {
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ DesktopEntryName)").arg(d->task.data()->classClass()));
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Query DesktopEntryName via class" << d->task.data()->classClass() << (int)(services.empty() ? 0 : 1);
#endif
        }

        // Try StartupWMClass
        if(services.empty()) {
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ StartupWMClass)").arg(d->task.data()->classClass()));
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Query StartupWMClass via class" << d->task.data()->classClass() << (int)(services.empty() ? 0 : 1);
#endif
        }

        // Try 'Name' - unfortunately this can be translated, so has a good chance of failing! (As it does for KDE's own "System Settings" (even in English!!))
        if(services.empty()) {
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ Name)").arg(d->task.data()->classClass()));
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Query Name via class" << d->task.data()->classClass() << (int)(services.empty() ? 0 : 1);
#endif
        }
        
        // Ok, absolute *last* chance, try matching via pid (but only if we have not already tried this!)...
        if(services.empty() && !triedPid) {
            services=getServicesViaPid(d->task.data()->pid());
        }
    }
    
    if(services.empty() && isStartupItem()) {
        // Try to match via desktop filename...
        if(!startup()->desktopId().isNull() && startup()->desktopId().endsWith(".desktop")) {
            if(startup()->desktopId().startsWith("/")) {
                d->launcherUrl=KUrl::fromPath(startup()->desktopId());
            #ifdef ICON_ENABLE_DBUS_DBUG
                ICN_LIB_L_DBUG << "Returning (from start-up)" << d->launcherUrl;
            #endif
                return d->launcherUrl;
            } else {
                QString desktopName=startup()->desktopId();
            
                if(desktopName.endsWith(".desktop")) {
                    desktopName=desktopName.mid(desktopName.length()-8);
                }
     
                services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ DesktopEntryName)").arg(desktopName));

                #ifdef ICON_ENABLE_DBUS_DBUG
                ICN_LIB_L_DBUG << "Start-up Query via applicationId" << desktopName << (int)(services.empty() ? 0 : 1);
                #endif
            }
        }
        
        // Try StartupWMClass
        if(services.empty() && !startup()->wmClass().isNull()) {
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ StartupWMClass)").arg(startup()->wmClass()));
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Start-up Query via StartupWMClass" << startup()->wmClass() << (int)(services.empty() ? 0 : 1);
#endif
        }
        
        // Try via name...
        if(services.empty() && !startup()->text().isNull()) {
            services = KServiceTypeTrader::self()->query("Application", QString("exist Exec and ('%1' =~ Name)").arg(startup()->text()));
#ifdef ICON_ENABLE_DBUS_DBUG
            ICN_LIB_L_DBUG << "Start-up Query via Name" << startup()->text() << (int)(services.empty() ? 0 : 1);
#endif
        }
    }

    if (!services.empty()) {
        d->launcherUrl=KUrl::fromPath(services[0]->entryPath());
    } /*else {
        // No desktop-file was found, so try to find at least the executable
        // usually it's the lower cased window class class, but if that fails let's trust it
        QString path = KStandardDirs::findExe(d->task.data()->classClass().toLower());
        if (path.isEmpty()) {
            path = KStandardDirs::findExe(d->task.data()->classClass());
        }

        if (!path.isEmpty()) {
            d->launcherUrl=KUrl::fromPath(path);
        }
    }*/
#ifdef ICON_ENABLE_DBUS_DBUG
    ICN_LIB_L_DBUG << "Returning " << d->launcherUrl;
#endif
    return d->launcherUrl;
}

void TaskItem::resetLauncherCheck()
{
    if(d->launcherUrl.isEmpty()) {
        d->checkedForLauncher=false;
    }
}

void TaskItem::close()
{
    if (!d->task) {
        return;
    }
    d->task.data()->close();
}

bool TaskItem::isActive() const
{
    if (!d->task) {
        return false;
    }
    return d->task.data()->isActive();
}

bool TaskItem::demandsAttention() const
{
    if (!d->task) {
        return false;
    }
    return d->task.data()->demandsAttention();
}

#ifdef ICON_ENABLE_DBUS_DBUG
void TaskItem::dump(QTextStream &str) const
{
    str << "<" << QString(metaObject()->className()).replace("::", "__")
        << " name=\"" << Qt::escape(name()) << "\" ptr=\"" << (void *)this 
        << "\" task=\"" << (void *)(d->task ? d->task.data() : 0)
        << "\" parentGroup=\"" << (void *)parentGroup() << "\" mgr=\"" << (void *)(parentGroup() ? parentGroup()->manager() : 0)
        << "\" id=\"" << id() << "\" isGrouped=\"" << isGrouped()
        << "\" itemType=\"" << itemType() << "\" isStartupItem=\"" << isStartupItem() << "\" isOnCurrentDesktop=\"" << isOnCurrentDesktop()
        << "\" isOnAllDesktops=\"" << isOnAllDesktops() << "\" desktop=\"" << desktop() << "\" isActive=\"" << isActive()
        << "\" demandsAttention=\"" << demandsAttention() << "\" launcherUrl=\"" << launcherUrl().url() << "\" />\n";
}
#endif

} // TaskManager namespace

#include "taskitem.moc"
